#include <stdio.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

#include "kamount.h"

typedef unsigned char hfs_byte_t[1];
typedef unsigned char hfs_word_t[2];
typedef unsigned long hfs_lword_t;

/*================ File-local data types ================*/

/*
 * The Macintosh Driver Descriptor Block
 *
 * On partitioned Macintosh media this is block 0.
 * We really only need the "magic number" to check for partitioned media.
 */
struct hfs_drvr_desc {
	hfs_word_t	ddSig;		/* The signature word */
	/* a bunch more stuff we don't need */
};

/*
 * The new style Mac partition map
 *
 * For each partition on the media there is a physical block (512-byte
 * block) containing one of these structures.  These blocks are
 * contiguous starting at block 1.
 */
struct new_pmap {
	hfs_word_t	pmSig;		/* Signature bytes to verify
					   that this is a partition
					   map block */
	hfs_word_t	reSigPad;	/* padding */
	hfs_lword_t	pmMapBlkCnt;	/* (At least in block 1) this
					   is the number of partition
					   map blocks */
	hfs_lword_t	pmPyPartStart;	/* The physical block number
					   of the first block in this
					   partition */
	hfs_lword_t	pmPartBlkCnt;	/* The number of physical
					   blocks in this partition */
	hfs_byte_t	pmPartName[32];	/* (null terminated?) string
					   giving the name of this
					   partition */
	hfs_byte_t	pmPartType[32];	/* (null terminated?) string
					   giving the type of this
					   partition */
	/* a bunch more stuff we don't need */
};

/*
 * The old style Mac partition map
 *
 * The partition map consists for a 2-byte signature followed by an
 * array of these structures.  The map is terminated with an all-zero
 * one of these.
 */
struct old_pmap {
	hfs_word_t		pdSig;	/* Signature bytes */
	struct 	old_pmap_entry {
		hfs_lword_t	pdStart;
		hfs_lword_t	pdSize;
		hfs_lword_t	pdFSID;
	}	pdEntry[42];
};


/*
 * The HFS Master Directory Block (MDB).
 *
 * Also known as the Volume Information Block (VIB), this structure is
 * the HFS equivalent of a superblock.
 *
 * Reference: _Inside Macintosh: Files_ pages 2-59 through 2-62
 *
 * modified for HFS Extended
 */
struct raw_mdb {
	hfs_word_t	drSigWord;	/* Signature word indicating fs type */
	hfs_lword_t	drCrDate;	/* fs creation date/time */
	hfs_lword_t	drLsMod;	/* fs modification date/time */
	hfs_word_t	drAtrb;		/* fs attributes */
	hfs_word_t	drNmFls;	/* number of files in root directory */
	hfs_word_t	drVBMSt;	/* location (in 512-byte blocks)
					   of the volume bitmap */
	hfs_word_t	drAllocPtr;	/* location (in allocation blocks)
					   to begin next allocation search */
	hfs_word_t	drNmAlBlks;	/* number of allocation blocks */
	hfs_lword_t	drAlBlkSiz;	/* bytes in an allocation block */
	hfs_lword_t	drClpSiz;	/* clumpsize, the number of bytes to
					   allocate when extending a file */
	hfs_word_t	drAlBlSt;	/* location (in 512-byte blocks)
					   of the first allocation block */
	hfs_lword_t	drNxtCNID;	/* CNID to assign to the next
					   file or directory created */
	hfs_word_t	drFreeBks;	/* number of free allocation blocks */
	hfs_byte_t	drVN[28];	/* the volume label */
	hfs_lword_t	drVolBkUp;	/* fs backup date/time */
	hfs_word_t	drVSeqNum;	/* backup sequence number */
	hfs_lword_t	drWrCnt;	/* fs write count */
	hfs_lword_t	drXTClpSiz;	/* clumpsize for the extents B-tree */
	hfs_lword_t	drCTClpSiz;	/* clumpsize for the catalog B-tree */
	hfs_word_t	drNmRtDirs;	/* number of directories in
					   the root directory */
	hfs_lword_t	drFilCnt;	/* number of files in the fs */
	hfs_lword_t	drDirCnt;	/* number of directories in the fs */
	hfs_byte_t	drFndrInfo[32];	/* data used by the Finder */
	hfs_word_t	drEmbedSigWord;	/* embedded volume signature */
	hfs_lword_t     drEmbedExtent;  /* starting block number (xdrStABN)
					   and number of allocation blocks
					   (xdrNumABlks) occupied by embedded
					   volume */
	hfs_lword_t	drXTFlSize;	/* bytes in the extents B-tree */
	hfs_byte_t	drXTExtRec[12];	/* extents B-tree's first 3 extents */
	hfs_lword_t	drCTFlSize;	/* bytes in the catalog B-tree */
	hfs_byte_t	drCTExtRec[12];	/* catalog B-tree's first 3 extents */
};

struct hfs_vol_head {
	hfs_word_t vhSigWord;
	hfs_byte_t vhUnknown[0x22];
	char vhVNLen;
  char vhVN[28];
};

/*================ File-local functions ================*/

hfs_lword_t l_lendian(hfs_lword_t in)
{
  char buf[4];
  char tmp[4];
  hfs_lword_t out;
  memcpy(buf,(char *)&in,4);
  tmp[0]=buf[3];
  tmp[1]=buf[2];
  tmp[2]=buf[1];
  tmp[3]=buf[0];
  memcpy((char *)&out,tmp,4);
  return out;
}

struct new_pmap *readnewpart(int fd)
{
	char buf[2];
	long p,i;
	struct new_pmap *map;
	p=lseek(fd,0,SEEK_CUR);
	for(i=p;i<0x5000;i++) {
		lseek(fd,i,SEEK_SET);
		read(fd,(char *)&buf,sizeof(buf));
		if(!memcmp(buf,"PM",2)) {
			//printf("Offset %08lx ",i);
			lseek(fd,i,SEEK_SET);
			map=(struct new_pmap *)malloc(sizeof(struct new_pmap));
			read(fd,(void *)map,sizeof(struct new_pmap));
			map->pmMapBlkCnt=l_lendian(map->pmMapBlkCnt);
			map->pmPyPartStart=l_lendian(map->pmPyPartStart);
			return map;
		}
	}
}

char *readnewname(int fd, hfs_lword_t bstart)
{
	unsigned long p;
	struct raw_mdb mdb;
	struct hfs_vol_head vh;
	char buf[3];
	char *ret=NULL;
	int l;
	p=lseek(fd,0,SEEK_CUR);
	lseek(fd,bstart*512,SEEK_SET);
	read(fd,(char *)&mdb,sizeof(struct raw_mdb));
	printf("%08lx\n",lseek(fd,0,SEEK_CUR));
	lseek(fd,bstart*512+0x400,SEEK_SET);
	read(fd,(char *)&vh,sizeof(struct hfs_vol_head));
	if(!memcmp(vh.vhSigWord,"BD",2)) {
		l=(int)vh.vhVNLen&0xff;
		vh.vhVN[l]=0;
		lseek(fd,p,SEEK_SET);
		printf("Veremos %d %s\n",l,vh.vhVN);
		ret=(char *)malloc(l+1);
		bzero(ret,l+1);
		memcpy(ret,vh.vhVN,l);
		return ret;
	} else {
		printf("No fue\n");
	}
	return NULL;
}
char *gethfsname(struct DEVICES *dev)
{
	int fd;
	int i;
	struct hfs_drvr_desc fblock;
	struct new_pmap *nmap;
	struct old_pmap *pmap;
	char *ret;
	fd=open(dev->device,O_RDONLY|O_NONBLOCK);
	read(fd,(void *)&fblock.ddSig,sizeof(fblock.ddSig));
	if(!memcmp(fblock.ddSig,"ER",sizeof(fblock.ddSig))) {
		printf("Nuevo fs\n");
		nmap=readnewpart(fd);
		printf("%08lx particiones\n",nmap->pmMapBlkCnt);
		printf("Nombre %s del tipo %s\n",nmap->pmPartName,nmap->pmPartType);
		for(i=1;i<nmap->pmMapBlkCnt;i++) {
			nmap=readnewpart(fd);
			printf("Nombre %s del tipo %s\n",nmap->pmPartName,nmap->pmPartType);
			if(!strcmp((char *)nmap->pmPartType,"Apple_HFS")) {
			  printf("Esta es la buena y empieza en bloque %08lx y byte %08lx\n",nmap->pmPyPartStart,nmap->pmPyPartStart*512);
			  ret=readnewname(fd,nmap->pmPyPartStart);
			  printf("El nombre fue %s\n",ret);
			  close(fd);
			  free(nmap);
			  return ret;
			}
		}
	} else {
		printf("Viejo???\n");
	}
	return NULL;
}
